;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "eglotx" "20260824.219"
  "Native LSP multiplexer for Eglot."
  '((emacs   "29.1")
    (eglot   "1.24")
    (jsonrpc "1.0.29"))
  :url "https://github.com/cxa/eglotx"
  :commit "2d5d2fbbd0ebbf2d07972a6b260ff6d9023a4586"
  :revdesc "2d5d2fbbd0eb"
  :keywords '("tools" "languages")
  :authors '(("CHEN Xian'an" . "xianan.chen@gmail.com"))
  :maintainers '(("CHEN Xian'an" . "xianan.chen@gmail.com")))
